@@ -22,6 +22,7 @@ from ..common import (
 # command line identation
 cmdind = "    "
 destdirs = set()
+catkin_srcs = []
 dpkg_info = query_dpkg_architecture()
 
 
@@ -133,14 +134,20 @@ def run(args, environ):
         return 0
 
     if args.clean:
-        for path, package in packages:
-            if package.name not in skipped_packages:
-                do_action(
-                    "clean", args, path, package, buildopts, install_prefix, environ
-                )
-        if args.builddir.is_dir():
-            log.info(f"Removing {args.builddir}")
-            shutil.rmtree(args.builddir, ignore_errors=True)
+        try:
+            for path, package in packages:
+                if package.name not in skipped_packages:
+                    do_action(
+                        "clean", args, path, package, buildopts, install_prefix, environ
+                    )
+            if args.builddir.is_dir():
+                log.info(f"Removing {args.builddir}")
+                shutil.rmtree(args.builddir, ignore_errors=True)
+        except KeyError as e:
+            log.error(str(e))
+            return 1
+        except subprocess.CalledProcessError as e:
+            return e.returncode or 1
         return 0
 
     sys.stdout.write(
@@ -213,6 +220,7 @@ def run(args, environ):
 
 def do_action(action, args, path, package, buildopts, install_prefix, environ):
     global destdirs
+    global catkin_srcs
     log = logging.getLogger("dhros")
     destdir = Path(
         get_build_var(package.name, "ROS_DESTDIR", args.destdir, env=environ)
@@ -220,10 +228,21 @@ def do_action(action, args, path, package, buildopts, install_prefix, environ):
     homedir = Path(
         get_build_var(package.name, "ROS_HOME", args.builddir / ".roshome", env=environ)
     ).resolve()
+    catkin_test_results_dir = Path(
+        get_build_var(
+            package.name,
+            "ROS_CATKIN_TEST_RESULTS_DIR",
+            args.builddir / package.name / "test_results",
+            env=environ,
+        )
+    ).resolve()
+    catkin_devel_dir = (args.builddir / "__devel__").resolve()
     relative_install_prefix = install_prefix.relative_to("/")
     destdirs.add(destdir)
     package_suffix = re.sub(r"[^A-Za-z0-9]", "_", package.name)
     build_type = package.get_build_type()
+    if build_type == "catkin":
+        catkin_srcs.append(str(path.resolve()))
     hook_vars = {
         "build_dir": shlex.quote(str((args.builddir / package.name).resolve())),
         "build_type": shlex.quote(build_type),
@@ -265,6 +284,8 @@ def do_action(action, args, path, package, buildopts, install_prefix, environ):
     prefixes = prefixes.split(":") if prefixes is not None else []
     if install_prefix != Path("/usr"):
         prefixes.append(str(install_prefix))
+    if build_type == "catkin":
+        prefixes.append(str(catkin_devel_dir))
     env["CMAKE_PREFIX_PATH"] = ":".join(
         prefixes + [str(d / relative_install_prefix) for d in destdirs]
     )
@@ -282,13 +303,17 @@ def do_action(action, args, path, package, buildopts, install_prefix, environ):
         if install_prefix != Path("/usr"):
             package_paths.append(str(install_prefix / "share"))
         env["ROS_PACKAGE_PATH"] = ":".join(
-            package_paths + [str(d / relative_install_prefix / "share") for d in destdirs]
+            package_paths
+            + catkin_srcs
+            + [str(d / relative_install_prefix / "share") for d in destdirs]
         )
 
     pythonpaths = env.get("PYTHONPATH", None)
     pythonpaths = pythonpaths.split(":") if pythonpaths is not None else []
     if build_type == "ament_cmake":
         pythonpaths += compute_ament_cmake_python_path(args.builddir / package.name)
+    if build_type == "catkin":
+        pythonpaths += compute_python_package_path(catkin_devel_dir)
     for d in destdirs:
         pythonpaths += compute_python_package_path(d / relative_install_prefix)
     if install_prefix != Path("/usr"):
@@ -329,9 +354,9 @@ def do_action(action, args, path, package, buildopts, install_prefix, environ):
     )
 
     if not execute_hook(
-        f"custom_{action}", package, hook_vars, env, dry_run=args.dry_run
-    ) and not execute_hook(
         f"custom_{action}_{build_type}", package, hook_vars, env, dry_run=args.dry_run
+    ) and not execute_hook(
+        f"custom_{action}", package, hook_vars, env, dry_run=args.dry_run
     ):
         cmdline = [f"dh_auto_{action}", f"--sourcedir={path}"]
         if build_type in ["ament_python"]:
@@ -357,7 +382,11 @@ def do_action(action, args, path, package, buildopts, install_prefix, environ):
                 if install_prefix != Path("/usr"):
                     cmdline.append(f"-DCMAKE_INSTALL_PREFIX={install_prefix}")
             if build_type == "catkin":
-                cmdline += ["-DCATKIN_BUILD_BINARY_PACKAGE=ON"]
+                cmdline += [
+                    "-DCATKIN_BUILD_BINARY_PACKAGE=ON",
+                    f"-DCATKIN_DEVEL_PREFIX={str(catkin_devel_dir)}",
+                    f"-DCATKIN_TEST_RESULTS_DIR={str(catkin_test_results_dir)}",
+                ]
             if build_type == "ament_cmake":
                 cmdline += ["-DAMENT_LINT_AUTO=OFF"]
                 if install_prefix == Path("/usr"):
@@ -391,10 +420,19 @@ def do_action(action, args, path, package, buildopts, install_prefix, environ):
         if not args.dry_run:
             subprocess.check_call(cmdline, env=env)
 
-            if action == "test" and build_type == "catkin" and package.name != "catkin":
-                sys.stdout.write(f"{cmdind}catkin_test_results\n")
+            if (
+                action == "test"
+                and build_type == "catkin"
+                and package.name != "catkin"
+                and catkin_test_results_dir.is_dir()
+            ):
+                cmdline = ["catkin_test_results"]
+                if args.verbose:
+                    cmdline.append("--verbose")
+                cmdline.append(str(catkin_test_results_dir))
+                sys.stdout.write(f"{cmdind}{shlex.join(cmdline)}\n")
                 sys.stdout.flush()
-                subprocess.check_call("catkin_test_results", env=env)
+                subprocess.check_call(cmdline, env=env)
 
     execute_hook(
         f"after_{action}_{build_type}", package, hook_vars, env, dry_run=args.dry_run
